(define-module (undone view)
  #:use-module (oop goops)
  #:export (<view> view-list sort-view filter-view print-item
                   get-list-for print-view view-doc define-view))

(define-syntax define-view
  (lambda (x)
    (syntax-case x (item)
      ((_ name doc print-item-form)
       (with-syntax ((view-name
                      (datum->syntax
                       x (symbol-append '< (syntax->datum #'name) '-view>)))
                     (item (datum->syntax x 'item)))
         #`(begin
             (define-class view-name (<view>))

             (define-method (view-doc (v view-name))
               doc)

             (define-method (print-item (v view-name) item)
               print-item-form)

             (set! view-list
                   (append view-list
                           (list (cons 'name view-name))))))))))

(define-generic sort-view)
(define-generic filter-view)
(define-generic print-item)
(define-generic get-list-for)
(define-generic print-view)
(define-generic view-doc)

(define-class <view> ()
  (todo-list #:init-value '() #:init-keyword #:todo-list))

(define view-list `((view . ,<view>)))

(define-method (sort-view (v <view>) l) l)

(define-method (print-item (v <view>) item)
  (format #t "~2d ~a~%"
          (cdr (assq 'id item))
          (cdr (assq 'content item))))

(define-method (filter-view (v <view>))
  (lambda (elm)
    (and (assq 'id elm) (assq 'content elm))))

(define-method (get-list-for (v <view>))
  (filter (filter-view v)
          (sort-view v (slot-ref v 'todo-list))))

(define-method (print-view (v <view>))
  (for-each
   (lambda (item) (print-item v item))
   (get-list-for v)))

(define-method (view-doc (v <view>))
  "Default view, show only the basics.")
